#include "RenderTraverser.h"
#include "glInterface.h"
#include "Timers.h"
#include "Camera.h"
#include "SceneEntity.h"


namespace CHCDemo
{

	
void TraversalStatistics::Reset()
{
	// reset statistics
	mNumTraversedNodes = 0;
	mNumQueryCulledNodes = 0;
	mNumFrustumCulledNodes = 0;
	mNumRenderedGeometry = 0;
	mNumRenderedTriangles = 0;

	mNumIssuedQueries = 0;
	mRenderTime = 0;
}



/******************************************************/
/*           RenderTraverser implementation           */
/******************************************************/


RenderTraverser::RenderTraverser():  
mVisibilityThreshold(0),
mBvh(NULL), 
mIsQueryMode(false),
mUseOptimization(true),
mFrameId(-1)
{
}


RenderTraverser::~RenderTraverser()
{
	mQueryHandler.DestroyQueries();
}


void RenderTraverser::InitFrame(Camera *camera, int currentFrameId)
{
	mFrameId = currentFrameId;
	mBvh->InitFrame(camera, currentFrameId);

	mStats.Reset();
	mQueryHandler.ResetQueries();
	mRenderState->mTexturesEnabled = false;
	EnqueueNode(mBvh->GetRoot());
}


void RenderTraverser::EnqueueNode(BvhNode *node)
{
	mBvh->CalcDistance(node);
	mDistanceQueue.push(node);
}


void RenderTraverser::TraverseNode(BvhNode *node)
{
	++ mStats.mNumTraversedNodes;

	if (node->IsVirtualLeaf())
	{
		//RenderBox(node->GetBox());
		mStats.mNumRenderedGeometry +=	mBvh->Render(node, mRenderState);
	}
	else 
	{
		// for non leafs this renders only the bounding volume (if the flag is set)
		BvhInterior *interior = static_cast<BvhInterior *>(node);

		EnqueueNode(interior->GetFront());
		EnqueueNode(interior->GetBack());
	}
}


void RenderTraverser::SetHierarchy(Bvh *bvh)
{
	mBvh = bvh;
	//DelQueries();
}


void RenderTraverser::SetRenderState(RenderState *state)
{
	mRenderState = state;
	//DelQueries();
}



void RenderTraverser::Switch2GLRenderState()
{
	// boolean used to avoid unnecessary state changes
	if (mIsQueryMode)
	{
		// switch back to rendermode		
		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
		glDepthMask(GL_TRUE);
		glEnable(GL_LIGHTING);
		mIsQueryMode = false;
	}
}


OcclusionQuery *RenderTraverser::IssueOcclusionQuery(BvhNode *node, bool wasVisible)
{
	++ mStats.mNumIssuedQueries;

	OcclusionQuery *query = mQueryHandler.RequestQuery();
	query->Reset();
	query->AddNode(node);

	query->BeginQuery();
	
	// if leaf and was visible => will be rendered anyway, thus we
	// can also test with the real geometry 
	/*if(node->IsLeaf() && wasVisible && mUseOptimization)
	{
		mNumRenderedGeometry += node->Render();
	}
	else
	{*/

	// change state so the bounding box gets not actually rendered on the screen
	Switch2GLQueryState();
	mBvh->RenderBoundingBox(node);
	Switch2GLRenderState();

	query->EndQuery();

	return query;
}


void RenderTraverser::Switch2GLQueryState()
{	
	// boolean used to avoid unnecessary state changes
	if (!mIsQueryMode)
	{
		glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
		glDepthMask(GL_FALSE);
		glDisable(GL_LIGHTING);
		mIsQueryMode = true;
	}
}



void RenderTraverser::PullUpVisibility(BvhNode *node)
{
	while(node && ! node->IsVisible())
	{
		node->SetVisible(true);
		node = node->GetParent();
	}
}


void RenderTraverser::SetVisibilityThreshold(int threshold)
{
	mVisibilityThreshold = threshold;
}


int RenderTraverser::GetVisibilityThreshold() const
{
	return mVisibilityThreshold;
}


/*
void RenderTraverser::RenderVisualization()
{
	mDistanceQueue.push(mHierarchyRoot);

	while(!	mDistanceQueue.empty())
	{
		HierarchyNode *node = mDistanceQueue.top();
		mDistanceQueue.pop();

		// identify previously visible nodes
		bool wasVisible = node->Visible() && (node->LastVisited() == mFrameID - 1);

		if(wasVisible)
			TraverseNode(node);
		else
		{
			// also render culled nodes
			glColor3f(1.0,0.0,0.0);
			node->RenderBoundingVolumeForVisualization();		
		}
	}
}*/


void RenderTraverser::SetUseOptimization(bool useOptimization)
{
	mUseOptimization = useOptimization;
	printf("using opt %d\n", mUseOptimization);
}


void RenderTraverser::RenderBox(const AxisAlignedBox3 &box)
{
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glDisable(GL_CULL_FACE);
	glColor3f(1.0f, 0.0f, 0.0f);

	glBegin(GL_QUADS);
	glVertex3f(box.Min().x, box.Max().y, box.Min().z);
	glVertex3f(box.Max().x, box.Max().y, box.Min().z);
	glVertex3f(box.Max().x, box.Min().y, box.Min().z);
	glVertex3f(box.Min().x, box.Min().y, box.Min().z);
	glEnd();

	glBegin(GL_QUADS);
	glVertex3f(box.Min().x, box.Min().y, box.Max().z);
	glVertex3f(box.Max().x, box.Min().y, box.Max().z);
	glVertex3f(box.Max().x, box.Max().y, box.Max().z);
	glVertex3f(box.Min().x, box.Max().y, box.Max().z);
	glEnd();

	glBegin(GL_QUADS);
	glVertex3f(box.Max().x, box.Min().y, box.Min().z);
	glVertex3f(box.Max().x, box.Min().y, box.Max().z);
	glVertex3f(box.Max().x, box.Max().y, box.Max().z);
	glVertex3f(box.Max().x, box.Max().y, box.Min().z);
	glEnd();

	glBegin(GL_QUADS);
	glVertex3f(box.Min().x, box.Min().y, box.Min().z);
	glVertex3f(box.Min().x, box.Min().y, box.Max().z);
	glVertex3f(box.Min().x, box.Max().y, box.Max().z);
	glVertex3f(box.Min().x, box.Max().y, box.Min().z);
	glEnd();

	glBegin(GL_QUADS);
	glVertex3f(box.Min().x, box.Min().y, box.Min().z);
	glVertex3f(box.Max().x, box.Min().y, box.Min().z);
	glVertex3f(box.Max().x, box.Min().y, box.Max().z);
	glVertex3f(box.Min().x, box.Min().y, box.Max().z);
	glEnd();

	glBegin(GL_QUADS);
	glVertex3f(box.Min().x, box.Max().y, box.Min().z);
	glVertex3f(box.Max().x, box.Max().y, box.Min().z);
	glVertex3f(box.Max().x, box.Max().y, box.Max().z);
	glVertex3f(box.Min().x, box.Max().y, box.Max().z);
	glEnd();

	glEnable(GL_CULL_FACE);

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}


void RenderTraverser::RenderFrustum()
{
	Vector3 ftl, ftr, fbl, fbr, ntl, ntr, nbl, nbr;
	mCamera->ComputePoints(ftl, ftr, fbl, fbr, ntl, ntr, nbl, nbr);

	glLineWidth(2);

	glBegin(GL_LINE_LOOP);
	glVertex3d(fbl.x, fbl.y, fbl.z);
	glVertex3d(fbr.x, fbr.y, fbr.z);
	glVertex3d(ftr.x, ftr.y, ftr.z);
	glVertex3d(ftl.x, ftl.y, ftl.z);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3d(nbl.x, nbl.y, nbl.z);
	glVertex3d(nbr.x, nbr.y, nbr.z);
	glVertex3d(ntr.x, ntr.y, ntr.z);
	glVertex3d(ntl.x, ntl.y, ntl.z);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3d(fbl.x, fbl.y, fbl.z);
	glVertex3d(ftl.x, ftl.y, ftl.z);
	glVertex3d(ntl.x, ntl.y, ntl.z);
	glVertex3d(nbl.x, nbl.y, nbl.z);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3d(fbr.x, fbr.y, fbr.z);
	glVertex3d(ftr.x, ftr.y, ftr.z);
	glVertex3d(ntr.x, ntr.y, ntr.z);
	glVertex3d(nbr.x, nbr.y, nbr.z);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3d(fbr.x, fbr.y, fbr.z);
	glVertex3d(fbl.x, fbl.y, fbl.z);
	glVertex3d(nbl.x, nbl.y, nbl.z);
	glVertex3d(nbr.x, nbr.y, nbr.z);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3d(ftr.x, ftr.y, ftr.z);
	glVertex3d(ftl.x, ftl.y, ftl.z);
	glVertex3d(ntl.x, ntl.y, ntl.z);
	glVertex3d(ntr.x, ntr.y, ntr.z);
	glEnd();
}



}
